"""
  This script contains a class mainly for preprocessing the data and return
"""

import os
import numpy as np
from collections import Counter
from khmerml.preprocessing.read_content import ReadContent
from khmerml.utils.file_util import FileUtil
from khmerml.utils.log import Log
from khmerml.utils.unicodesplit import UnicodeSplit

class Preprocessing(object):
  """"
    This is the main class
  """
  def __init__(self, **kwargs):
    self.kwargs = kwargs

  def loading_data(self, folders, feature_choice, fq_type, threshold):
    """ Loading the data from txt files in folders """
    content = ReadContent(**self.kwargs)
    _words_articles, _all_words = content.load_content(folders)
    _temp_all_words = _all_words
    _temp_all_words.append(feature_choice)
    _temp_all_words.append(threshold)


    if feature_choice == 'doc_freq':
      _tfidf = self.doc_frequency(_words_articles, _all_words, feature_choice, fq_type, threshold)
    return _tfidf

  def doc_frequency(self, words_articles, all_words, feature_choice, fq_type, threshold):
    """
      This method aims at feature selection based
      on terms appearing in the articles.
      Select only terms that appear in more threshold of articles
    """
    _preq_words = dict() # words prequency with class
    _frequency = dict() # words prequency
    _label_match = dict()
    _label_count = 0
    for label in words_articles.keys(): # each class
      _label_match[label] = _label_count
      _label_count += 1
      if label not in _preq_words:
        _preq_words[label] = dict()
      for doc in words_articles[label]: # all words in each class
        for word in doc:
          if word in _frequency:
            _frequency[word] += 1
          else:
            _frequency[word] = 1
          if word in _preq_words[label]:
            _preq_words[label][word] += 1
          else:
            _preq_words[label][word] = 1
    # print('label', _label_match)

    _tfidf_mat = []
    _selected_words = []
    if fq_type == 'all':
      _selected_words = self.fq_all(_frequency, threshold)
    elif fq_type == 'class':
      _selected_words = self.fq_by_class(_preq_words, threshold)
    else:
      _selected_words = self.fq_by_top(_preq_words, threshold)

    _selected_words = list(set(_selected_words))
    _bag_of_words = self.compute_feature_matrix(_selected_words,\
                        words_articles, _label_match, feature_choice, threshold)
    # Save the dictionary for single document prediction
    path = self.kwargs['bag_of_words'] if ('bag_of_words' in self.kwargs) else 'data/bag_of_words'
    FileUtil.create_folder(path)
    FileUtil.save_pickle_dataset(path + '/'+feature_choice+'_'+str(threshold)+'.pickle', _selected_words)
    FileUtil.save_pickle_dataset(path + '/label_match.pickle', _label_match)
    return _bag_of_words


  def compute_feature_matrix(self, word_in_dic, text_in_articles, label_match, feature_choice, threshold):
    """
      Computing the feature matrix
    """
    mat = [] # feature matrix
    for label in text_in_articles.keys(): # each class
      for doc in text_in_articles[label]: # all words in each class
        row = []
        dic = Counter(doc)
        for word in word_in_dic:
          # each term or feature in article to be considered
          row.append(dic[word])# Adding to row
        row.append(label_match[label]) # adding label
        mat.append(row)

    _directory_name = FileUtil.join_path(self.kwargs['dataset'] if ('passion' in self.kwargs) else 'data/matrix' )
    FileUtil.create_folder(_directory_name)
    self.write_mat(_directory_name, feature_choice, threshold, mat)
    # returning feature matrix
    return mat

  def write_mat(self, path_to_file, feature_choice, threshold, feature_mat):
    """
      Method to write matrix into a file
    """
    FileUtil.create_folder(path_to_file)
    _mat_file = open(path_to_file + '/' + feature_choice+'_'+str(threshold)+'.csv', "w+")

    for _feature in feature_mat:
      _temp = str(_feature)
      _temp = _temp.strip(']')
      _temp = _temp.strip('[')
      _mat_file.write(_temp + '\n')

    _mat_file.close()

  def loading_single_doc(self, document, feature_choice, threshold):
    """
      Loading single document for prediction
    """
    content = ReadContent(**self.kwargs)
    if feature_choice == 'doc_freq':
      dic_load = FileUtil.load_pickle((self.kwargs['bag_of_words'] if ('bag_of_words' in self.kwargs) else 'data/bag_of_words') + \
                                      '/'+feature_choice+'_'+str(threshold)+'.pickle')
    _new_words = None
    is_unicode = self.kwargs.get('is_unicode', False)
    if is_unicode  :
      _new_words = UnicodeSplit().unicode_split(document)
    else:
      _new_words = content.remove_stopword(document)
      _new_words = content.stemming_words(_new_words)
    words = Counter(_new_words)# Count the frequency of each term
    row = []
    for word in dic_load:
      # each term or feature in article to be consider
      row.append(words[word])# Adding to row
    row.append(0)
    return row

  def fq_all(self,frequency, threshold):
    _selected_words = []
    for word in frequency:
      if frequency[word] >= threshold:
        # Consider only terms which appear in
        # documents more than a threshold
        _selected_words.append(word)
    Log('fq_all_selected_word.log').log(_selected_words)
    return _selected_words

  def fq_by_class(self,frequency_class, threshold):
    _selected_words = []
    for label,words in frequency_class.items():
      for word in frequency_class[label]:
        if words[word] >= threshold:
          _selected_words.append(word)

    Log('fq_by_class_selected_word.log').log(_selected_words)
    return _selected_words

  def fq_by_top(self,frequency_class, top):
    _selected_words = []
    for label, words in frequency_class.items():
      print('label',label)
      most_common = Counter(words).most_common(top)
      for word in most_common:
        _selected_words.append(word[0])

    Log('fq_by_class_top.log').log(_selected_words)
    return _selected_words


  def normalize_dataset(self, dataset):
    """ Transform data frequency to categorical data
    """
    feature_15 = dataset
    # % 0 : mean, 1 : mid point
    choice = 0
    # Every Attribute, determind the best information gain/gini
    for col_index in range(feature_15.shape[1] - 1):
      # Sort dataset following descendant label
      # dataset[:, -1].argsort # Sort the last field (column)
      # dataset = dataset[dataset[:,1].argsort(kind='mergesort')]
      feature = dataset[:, col_index]
      sort_feature = np.sort(feature)
      if choice == 1:
          index = np.floor(len(sort_feature)/2).astype('int')
          value = sort_feature[index]
      elif choice == 0:
          value = np.average(sort_feature)

      feature[sort_feature < value] = 0
      feature[sort_feature >= value] = 1

      dataset[:, col_index] = feature

    return dataset




"""
    This script aims at reading the articles from the folders
"""

import os
import nltk
from nltk.corpus import stopwords
from nltk.stem.lancaster import LancasterStemmer
from khmerml.utils.file_util import FileUtil
from khmerml.utils.unicodesplit import UnicodeSplit

class ReadContent(object):
  """"
    This is the main class
  """

  def __init__(self, **kwargs):
    self.kwargs = kwargs

  def load_content(self, dname):
    """ Below scripts are to get content from all files in each folders """
    _save_path = os.getcwd()
    _directory_name =  FileUtil.join_path(dname)
    # Directory containing the files
    # path_to_dataset = FileUtil.dataset_path(config, filename)
    _words_all_articles = dict()
    for folder in os.listdir(_directory_name):
      _words_each_articles = []   # List for storing all words in articles
      if folder == '.DS_Store' : continue
      for files in os.listdir(_directory_name+'/'+folder):
        if files.endswith(".txt"):
          _read_text = open(_directory_name+'/'+folder+'/'+files, "rU",\
                          encoding="utf-8", errors="surrogateescape")
          # Open file for reading
          _lines = _read_text.read()# Read content from file
          _new_words = []
          is_unicode = self.kwargs.get('is_unicode', False)
          if is_unicode  :
            _new_words = UnicodeSplit().unicode_split(_lines)
          else:
            _new_words = self.remove_stopword(_lines)
            _new_words = self.stemming_words(_new_words)
          _words_each_articles.append(_new_words)# Adding list to list
      _words_all_articles[folder.lower()] = _words_each_articles
      os.chdir(_save_path)  # Moving directory to the saved path
    _all_words = self.merge_list_content(_words_all_articles)

    return _words_all_articles, _all_words

  def remove_stopword(self, text):
    """ Removing stop words
    """
    _stop_words = set(stopwords.words('english')) # Stop words from English language
    _new_words = [i for i in  nltk.word_tokenize(text)\
                if i not in _stop_words]# Remove stop words
    # remove all tokens that are not alphabetic (non-english word)
    _new_words = [word for word in _new_words if word.isalpha()]
    return _new_words

  def stemming_words(self, text):
    """" Stemming the article"""

    _stemmer = LancasterStemmer()
    # _ignore_words = ['?', '$', '-', ',', '.', ';', ':', '``', '\'s', '\udc94', '[', ']', '%']
    _ignore_words = ['?', '$']
    _word = [_stemmer.stem(w.lower()) for w in text \
                            if w not in _ignore_words]# Stemming the words

    return _word

  def merge_list_content(self, list_of_list):
    """ This method is for merging list contents in list into one list """
    _all_words = []
    for counter in list_of_list:
      for innercounter in list_of_list[counter]:
        _all_words = _all_words + innercounter

    return list(set(_all_words))



import os
import warnings
from os.path import join
from numpy.distutils.core import setup


def configuration(parent_package='', top_path=None):
  from numpy.distutils.misc_util import Configuration
  #from numpy.distutils.system_info import get_info, BlasNotFoundError
  #import numpy
  libraries = []
  if os.name == 'posix':
      libraries.append('m')
  config = Configuration('khmerml', parent_package, top_path)
  # submodules which do not have their own setup.py
  # we must manually add sub-submodules & tests
  # config.add_subpackage('examples')
  config.add_subpackage('algorithms')
  config.add_subpackage('utils')
  # config.add_subpackage('naive_bayes')
  config.add_subpackage('preprocessing')
  return config

if __name__ == '__main__':
  setup(**configuration(top_path='').todict())




import numpy as np

class Tree(object):
  """
    Tree class is use for building Trees
    and use in DecisionTree Algorithm
  """

  def __init__(self, **kwargs):
    self.feature = None
    self.label = None
    self.n_samples = None
    self.gain = None
    self.left = None
    self.right = None
    self.threshold = None
    self.depth = 0
    self.max_depth = kwargs['max_depth']
    self.kwargs = kwargs

  def build(self, features, target, criterion=None):
    """
      Build a decision tree
    """
    if criterion is None:
      criterion = 'gini'
    # Number of rows of sample
    self.n_samples = features.shape[0]
    # End if all data are same class
    if len(np.unique(target)) == 1:
      self.label = target[0]
      return
    # check for max depth
    if self.depth > self.max_depth:
      return
    best_gain = 0.0
    best_feature = None
    best_threshold = None

    # Classification trees: the most common class, regression tree in the sample: the average in the sample
    if criterion in {'gini', 'entropy', 'error'}:
      # Find the label having high occurences
      self.label = max([(c, len(target[target == c])) for c in np.unique(target)],\
          key=lambda x: x[1])[0]
    else:
        self.label = np.mean(target)
    # Determine node impurity
    impurity_node = self._calc_impurity(criterion, target)
    # Every Attribute, determind the best information gain/gini
    for col_index in range(features.shape[1]):
      # Test split to get the best split
      info_gain, feature_index, threshold = \
      self.test_split(criterion, impurity_node, col_index, features, target)

      # Maximize information gain
      if info_gain > best_gain:
        best_gain = info_gain
        best_feature = feature_index
        best_threshold = threshold
    # No best threshold found,
    # terminate tree building for the current branch
    if best_threshold is None:
      return

    # Start develop left, then right branch
    self.feature = best_feature
    self.gain = best_gain
    self.threshold = best_threshold
    self._divide_tree(features, target, criterion)

  def test_split(self, criterion, impurity_node, index, features, target):
    """
      Split a dataset based on an attribute and an attribute value
    """
    best_gain = 0.0
    best_feature = None
    best_threshold = None
    # Remove redundant features
    feature_level, occurence_freq = np.unique(features[:, index], return_counts=True)
    #elements, occurence_labels = numpy.unique(target_class, return_counts=True)
    # More than one features found
    # Split tree based on "Middle Point"
    if occurence_freq is not None:
      # Pass over the element having frequency zero
      if not feature_level.any():
        return best_gain, best_feature, best_threshold
      if feature_level.shape[0] < 2:
        thresholds = (feature_level[:-1]) / 2.0
      else:
        thresholds = (feature_level[:-1] + feature_level[1:]) / 2.0

      # Explore
      for threshold in thresholds:
        target_l = target[features[:, index] <= threshold]
        impurity_l = self._calc_impurity(criterion, target_l)
        n_l = float(target_l.shape[0]) / self.n_samples

        target_r = target[features[:, index] > threshold]
        impurity_r = self._calc_impurity(criterion, target_r)
        n_r = float(target_r.shape[0]) / self.n_samples

        # (information gain): IG = node - (left + right)
        info_gain = impurity_node - (n_l * impurity_l + n_r * impurity_r)

        # Maximize information gain
        if info_gain > best_gain:
          best_gain = info_gain
          best_feature = index
          best_threshold = threshold
    return best_gain, best_feature, best_threshold

  def _divide_tree(self, features, target, criterion):

    features_l = features[features[:, self.feature] <= self.threshold]
    target_l = target[features[:, self.feature] <= self.threshold]
    #self.left = Tree(self.max_depth)
    self.left = Tree(**self.kwargs)
    self.left.depth = self.depth + 1
    self.left.build(features_l, target_l, criterion)

    features_r = features[features[:, self.feature] > self.threshold]
    target_r = target[features[:, self.feature] > self.threshold]
    self.right = Tree(**self.kwargs)
    self.right.depth = self.depth + 1
    self.right.build(features_r, target_r, criterion)

  def _calc_impurity(self, criterion, target):
    # Number of classes, and
    # Number of sample classes
    unique_target = np.unique(target)
    n_rows = target.shape[0]

    if criterion == 'gini':
      return self._gini(target, unique_target, n_rows)
    elif criterion == 'entropy':
      return self._entropy(target, unique_target, n_rows)
    elif criterion == 'error':
      return self._error(target, unique_target, n_rows)
    elif criterion == 'mse':
      return self._mse(target)
    else:
      return self._gini(target, unique_target, n_rows)

  # Gini impurity
  def _gini(self, target, n_classes, n_samples):
    """
      Calculate the Gini index
    """
    gini_index = 1.0
    gini_index -= sum([(float(len(target[target == label])) / float(n_samples)) ** 2.0 for label in n_classes])
    return gini_index

  # Entropy
  def _entropy(self, target, n_classes, n_samples):
    """
      Calculate the entropy of a dataset.
      The only parameter of this function is the target_col parameter which specifies the target column
    """
    entropy = 0.0
    # elements: each unique label in the label array
    # counts: the number of times each unique label appears in label array
    for label in n_classes:
      prob = float(len(target[target == label])) / n_samples
      if prob > 0.0:
        entropy -= prob * np.log2(prob)
    return entropy

  def _error(self, target, n_classes, n_samples):
    """
      Classification error
    """
    return 1.0 - max([len(target[target == label]) / n_samples for label in n_classes])

  def _mse(self, target):
    """
      Mean square error
    """
    y_hat = np.mean(target)
    return np.mean((target - y_hat) ** 2.0)

  def prune(self, method, limit_depth, min_criterion, n_samples):
    """
      Prune decision tree
    """

    if self.feature is None:
      return

    self.left.prune(method, limit_depth, min_criterion, n_samples)
    self.right.prune(method, limit_depth, min_criterion, n_samples)

    pruning = False

    # Pruning decision, Leaf
    if method == 'impurity' and self.left.feature is None and self.right.feature is None:
      if (self.gain * float(self.n_samples) / n_samples) < min_criterion:
        pruning = True
    elif method == 'depth' and self.depth >= limit_depth:
      pruning = True

    # Pruning suppresses over learning
    if pruning is True:
      self.left = None
      self.right = None
      self.feature = None


  def _predict(self, X_test):
    """
      Prediction of a new/unseen query instance.
    """
    # print('self.feature',self.feature,X_test)
    # Node
    if self.feature is not None:
      if X_test[self.feature] <= self.threshold:
        return self.left._predict(X_test)
      else:
        return self.right._predict(X_test)
    else: # Leaf
      # print(' self.label', self.label)
      return self.label

  def show_tree(self, depth, cond):
    """
      Display the structure of built tree from learning process
    """
    base = '    ' * depth + cond
    if self.feature != None: # Node
      print(base + 'if X[' + str(self.feature) + '] <= ' + str(self.threshold))
      self.left.show_tree(depth+1, 'then ')
      self.right.show_tree(depth+1, 'else ')
    else: # Leaf
      print(base + '{value: ' + str(self.label) + ', samples: ' + str(self.n_samples) + '}')



import os
import subprocess
import time

class KhmerSegment:
	def __init__(self):
		self.PATH = '/tmp/km-5tag-seg-1.0'

	def add_to_file(self, text):
		seconds = str(int(time.time() * 1000000))
		file_name = seconds + '.txt'
		file = self.PATH + '/sample/' + file_name
		f = open(file, 'w')
		f.write(text)
		f.close()
		return file_name

	def read_file(file):
		f = open(file, 'r')
		content = f.read()
		f.close()
		return content

	def remove_file(file):
		try:
			os.remove(file)
		except OSError:
			pass

	def run_command(command):
		FNULL = open(os.devnull, 'w')
		return subprocess.call(command, stdout = FNULL, stderr=subprocess.STDOUT, shell=True)




import numpy
import ntpath
from numpy import dot

from khmerml.utils.file_util import FileUtil
from khmerml.algorithms.base import Base
from khmerml.algorithms.neural_network.neural_network_base import NeuralNetworkBase


class NeuralNetwork(Base, NeuralNetworkBase):
  """
    Back-Propagation Neural Networks Written in Python.
  """

  def __init__(self, **kwargs):
    # Random range of initial weight
    self.random_state = kwargs['random_state']
    # Number of hidden layers
    self.hidden_layer_sizes = kwargs['hidden_layer_sizes']
    # Learning rate: speed up the learning process
    # to reach converge point
    self.learning_rate = kwargs['learning_rate']
    # Maximum numbers of learning iterations (times)
    self.max_iter = kwargs['max_iter']
    # Momentum
    self.momentum = kwargs['momentum']
    # activiation function
    self.activation = kwargs['activation']
    # Initial synaptic weight
    self.synaptic_weights = None
    # Change output
    self.change_output_layouts = None
    # Call super
    NeuralNetworkBase.__init__(self, **kwargs)

  def load_model(self):
    """
      Load train model from file
    """
    try:
      head, tail = ntpath.split(self.kwargs['train_model'])
      self.train_model = FileUtil.load_model(head+ '/neural_network_' +tail)
    except IOError as error:
      raise Exception(error)

    return self.train_model

  def save_model(self, model):
    """
      Load train model from file
    """
    try:
      head, tail = ntpath.split(self.kwargs['train_model'])
      FileUtil.save_model(head+ '/neural_network_' +tail, model)
    except IOError as error:
      print(error)

  def __calculate_activation_output(self, activation, output_neurons):
    """
      param activation: The activation function to be used. Can be
      logistic or tanh
    """
    if activation == 'logistic':
      return self.activation_sigmoid_exp(output_neurons)
    elif activation == 'tanh':
      return self.activation_tanh(output_neurons)
    elif activation == 'relu':
      return self.activation_relu(output_neurons)
    else:
      return None


  def __calculate_error(self, activation, output_neurons, delta_error):
    """
      param activation: The activation function to be used. Can be
      (logistic or tanh)
    """
    if activation == 'logistic':
      return self.derivative_sigmoid(output_neurons, delta_error)
    elif activation == 'tanh':
      return self.derivative_tanh(output_neurons, delta_error)
    elif activation == 'relu':
      return self.derivative_relu(output_neurons, delta_error)
    else:
      return None


  def feedforward(self, input_vector):
    """
      The neural network thinks.
    """
    # Output layer
    output_neurons = None
    output_by_layers = {}
    # Iterate through layers
    for layer_index, synaptic_weight in enumerate(self.synaptic_weights):
      #if layer_index == 0:
      if output_neurons is None:
        product = dot(synaptic_weight.T, input_vector)
      else:
        #product = dot(output_neurons.T, synaptic_weight)
        product = dot(synaptic_weight.T, output_neurons)
      # Calculate output based on given activation function
      output_neurons = self.__calculate_activation_output(self.activation, product)
      # Update nodes of the layer
      output_by_layers[layer_index] = output_neurons
    return output_by_layers


  def backpropagate(self, input_vector, output_by_layers, target_vector):
    """
      Back propagate of learning process
    """
    # Length of the list
    layers_len = len(output_by_layers)
    delta_error = None
    error = None
    delta_error_layers = {}

    # Loop backward from length of layers
    for index in reversed(range(layers_len)):
      # Calculate the error from output layer
      # (The difference between the desired output and the predicted output).
      if delta_error is None:
        output = output_by_layers[index]
        delta_error = -(target_vector - output)
        #output_error = target_vector - output
      else:
        # Calculate the error for layer 1 (By looking at the weights in layer 1,
        # we can determine by how much layer 1 contributed to the error in layer 2).
        synaptic_weight = self.synaptic_weights[index+1]
        output = output_by_layers[index]
        delta_error = numpy.dot(synaptic_weight, error)

      # Determine error between expected output and that produces from learning process
      error = self.__calculate_error(self.activation, output, delta_error)

      # Keep tracking of delta error of each layer
      delta_error_layers[index] = error

    # Determine how much weight to be adjusted of each synaptic weight
    for indice in reversed(range(layers_len)):
      if indice - 1 >= 0:
        previous_layer_output = output_by_layers[indice-1]
      else:
        previous_layer_output = input_vector

      # Calculate how much to adjust the weights by
      # error = delta_error_layers[indice]
      previous_output = numpy.atleast_2d(previous_layer_output)
      delta = numpy.atleast_2d(delta_error_layers[indice])
      # update the weights connecting hidden to output, change == partial derivative
      change = numpy.dot(previous_output.T, delta)
      # weight adjustment
      weight_adjustment = self.learning_rate * change
      # Adjust the weights, and Momentum addition
      self.synaptic_weights[indice] -= weight_adjustment + \
                                          self.change_output_layouts[indice] * self.momentum
      self.change_output_layouts[indice] = weight_adjustment

    return self.synaptic_weights


  def start_learning(self, input_vector, target_vector):
    """
      Learning process of input data and model.
      Multiplayer perceptron is adopted in this algorithm.
    """
    # One iteration of learning process composes of feedforward and backpropagation
    # Start learning feedforward following by backward propagation
    output_by_layers = self.feedforward(input_vector)
    self.synaptic_weights = self.backpropagate(input_vector, output_by_layers, target_vector)


  def train(self, train_sample):
    """
    """
    y_train = train_sample[:, -1]
    X_train = numpy.delete(train_sample, -1, 1)
    # Get label from dataset
    # Convert label to array of vector
    #label_vector = dataset_matrix[:, -1]
    y_target = self.label_to_matrix(y_train)
    # Number of layers
    n_features = len(X_train[0])
    # Number of output labels
    n_labels = len(y_target[0])
    # Initialize random weight (UI,  UH,  UO), iteration = 0
    # Create matrix of n_inputs_neuron as row and n_neurons as column
    self.synaptic_weights, self.change_output_layouts = self.make_weight_matrix(n_features=n_features, \
    n_labels=n_labels, hidden_layer_sizes=self.hidden_layer_sizes)

    for idx in range(self.max_iter):
      for index, input_vector in enumerate(X_train):
          self.start_learning(input_vector, y_target[index])
      # learning rate decay
      rate_decay = 0.0001
      self.learning_rate = self.learning_rate * (self.learning_rate / (self.learning_rate + (self.learning_rate * rate_decay)))

    # Save model
    self.save_model(self.synaptic_weights)
    # model
    return self.synaptic_weights


  def predict(self, model, test_sample):
    """
      Make prediction
    """
    # Remove class from X_test
    X_test = numpy.delete(test_sample, -1, 1)
    predictions = {}
    for index, vector in enumerate(X_test):
      # Start predicting one input per time
      result_dict = self._predict(model, vector)
      best_class = numpy.argmax(result_dict[len(result_dict) -1])
      # Add result to dict (key, value)
      # The expected output is at the end of list
      predictions[index] = best_class
      #predictions[index] = result_dict[len(result_dict) -1]
    return predictions


  def _predict(self, model, input_vector):
    """ The neural network thinks.
    """
    # Output layer
    output_neurons = None
    output_by_layers = {}
    # Iterate through layers
    for layer_index, synaptic_weight in enumerate(model):
      #if layer_index == 0:
      if output_neurons is None:
        product = dot(synaptic_weight.T, input_vector)
      else:
        product = dot(synaptic_weight.T, output_neurons)
      # Calculate output based on given activation function
      output_neurons = self.__calculate_activation_output(self.activation, product)
      # Update nodes of the layer
      output_by_layers[layer_index] = output_neurons
    return output_by_layers



import random
import numpy

from khmerml.algorithms.bayes.naive_bayes import NaiveBayes
from khmerml.algorithms.decisiontree.decision_tree import DecisionTree
from khmerml.algorithms.neural_network.neural_network import NeuralNetwork
from khmerml.utils.file_util import FileUtil

class MachineLearning(object):

  def __init__(self, *args, **kwargs):
      self.kwargs = kwargs

  def NiaveBayes(self):
    """
      Naivae Bayes Algorithm
    """
    return NaiveBayes(**self.kwargs)

  # Decision Tree Algorithm
  def DecisionTree(self,criterion='gini', prune='depth', max_depth=3, min_criterion=0.05):
    """
      DecisionTree Algorithm
    """
    self.kwargs['criterion'] = criterion
    self.kwargs['prune'] = prune
    self.kwargs['max_depth'] = max_depth
    self.kwargs['min_criterion'] = min_criterion
    return DecisionTree(**self.kwargs)


  def NeuralNetwork(self, hidden_layer_sizes=(100,), learning_rate=0.5, momentum=0.2, random_state=0, \
      max_iter=200, activation='tanh'):
    """
      NeuralNetwork Algorithm
    """
    self.kwargs['hidden_layer_sizes'] = hidden_layer_sizes
    self.kwargs['learning_rate'] = learning_rate
    self.kwargs['max_iter'] = max_iter
    self.kwargs['momentum'] = momentum
    self.kwargs['activation'] = activation
    self.kwargs['random_state'] = random_state
    return NeuralNetwork(**self.kwargs)


  def classify_dataset_by_class(self, dataset):
    """
      Classify dataset by class
    """
    dataset_by_class = {}
    for _, subset in enumerate(dataset):
      vector = subset
      if vector[-1] not in dataset_by_class:
        dataset_by_class[vector[-1]] = []
      dataset_by_class[vector[-1]].append(vector)
    return dataset_by_class

  def split_dataset(self, dataset, sample_by_class, use_numpy=True):
    """
      Split data set for training and testing
    """
    if use_numpy:
      return self.train_test_split(dataset, n_test_by_class=2)
    else:
      return self.split_dataset_py(dataset, sample_by_class)

  def split_dataset_py(self, dataset, sample_by_class):
    """
      Split data set for training and testing
    """
    dataset_by_class = self.classify_dataset_by_class(dataset)
    test_set = []
    no_train_set = 0
    for _, label in enumerate(dataset_by_class):
      no_train_set = 0
      subset = dataset_by_class[label]
      while no_train_set < sample_by_class:
        index_subset = random.randrange(len(subset))
        test_set.append(dataset_by_class[label].pop(index_subset))
        no_train_set = no_train_set + 1
    return [dataset_by_class, test_set]


  def train_test_split(self, dataset, n_test_by_class=2):
    """
      Split data set for training and testing
    """
    # Sort dataset following descendant label
    # dataset[:, -1].argsort # Sort the last field (column)
    # dataset = dataset[dataset[:,1].argsort(kind='mergesort')]
    sorted_dataset = dataset[dataset[:, -1].argsort()]
    # Get unique labels
    y_all = numpy.unique(sorted_dataset[:, -1])
    # Random row indices
    random_row_indices = None
    for _, label_val in enumerate(y_all):
      # Find row index where label value equals to given label_value
      search_indices = numpy.where(sorted_dataset[:, -1] == label_val)
      # Selected number of test sample equals to number of test sample
      random_idx = numpy.random.randint(search_indices[0][0], search_indices[0][-1], size=n_test_by_class)
      #mask = numpy.random.choice([False, True], lenght, p=[probability, 1-probability])
      # Merge array random row into one array
      if random_row_indices is None:
        random_row_indices = random_idx
      else:
        random_row_indices = numpy.hstack((random_row_indices, random_idx))
    # Split
    # Create mask boolean array
    mask = numpy.ones(len(sorted_dataset), dtype=bool)
    mask[random_row_indices, ] = False
    # instead of a[b] you could also do a[~mask]
    X_test, X_train = sorted_dataset[random_row_indices], sorted_dataset[mask]
    return X_train, X_test


  def accuracy(self, predictions, test_set):
    """
      Get Accuracy
    """
    correct = 0
    for index, _ in enumerate(test_set):
      if test_set[index][-1] == predictions[index]:
        correct += 1
    return round((correct / float(len(test_set))) * 100.0, 2)


  def to_label(self,label_values, file):
    label_load = FileUtil.load_pickle(file)
    result = []
    for label_value in label_values:
      for label, value in label_load.items():
        if value == label_value:
          result.append(label)
    return result




import copy
import ntpath
import numpy as np

from khmerml.algorithms.bayes.bayes_base import BayesBase
from khmerml.utils.file_util import FileUtil
from khmerml.algorithms.base import Base

class NaiveBayes(Base, BayesBase):
  """
    Naive Bayes class
  """

  def __init__(self, **kwargs):
    # BayesBase.__init__(**kwargs)
    self.kwargs = kwargs
    self.train_model = {}
    self.predictions = []

  def load_model(self):
    """
      Load train model from file
    """
    try:
      head, tail = ntpath.split(self.kwargs['train_model'])
      self.train_model = FileUtil.load_model(head+ '/naive_bayes_' +tail)
    except IOError as error:
      raise Exception(error)

    return self.train_model

  def save_model(self, model):
    """
      Load train model from file
    """
    try:
      head, tail = ntpath.split(self.kwargs['train_model'])
      FileUtil.save_model(head+ '/naive_bayes_' +tail, model)
    except IOError as error:
      print(error)

  def train(self, X_train):
    """
      Train model
    """
    prioris = self.calculate_priori(X_train)
    likelihoods = self.calculate_likelihood(X_train)

    train_model = {}
    for label, likelihood in likelihoods.items():

      # Get priori for corresponding class
      priori = prioris[label]
      if label not in train_model:
        train_model[label] = []

      # Push priori and likelihood of each label to stack
      train_model[label].append(priori)
      train_model[label].append(likelihood)

    self.train_model = train_model
    self.save_model(train_model)
    return train_model

  def predict(self, model, X_test):
    """
      Make prediction
    """

    predictions = []
    X_test = np.delete(X_test, -1, 1)

    for index in range(X_test.shape[0]):
    #for subset in test_sample:
      # remove label from test dataset
      #test_vector = np.delete(X_test[index], -1, 1)
      test_vector = X_test[index]
      _, label = self.calculate_posteriori(model, test_vector)
      predictions.append(label)
    self.predictions = predictions

    return predictions



import os
import subprocess
import time
from .khmersegment import KhmerSegment

class UnicodeSplit:
  def __init__(self):
    pass

  # Process Khmer Segmentation
  def unicode_split(self,text):
    segment = KhmerSegment()
    file_name = segment.add_to_file(text)

    # Run command to convert the sentence into separated words
    command = 'cd ' + segment.PATH + ' && ./km-5tag-seg-test.sh model/km-5tag-seg-model sample/' + file_name + ' sample-out/'
    KhmerSegment.run_command(command)

    input_file = segment.PATH + '/sample/' + file_name
    output_file_w = segment.PATH + '/sample-out/' + file_name + '.w'
    output_file_c = segment.PATH + '/sample-out/' + file_name + '.c'

    # Read temporary file content
    result = KhmerSegment.read_file(output_file_w).split(' ')
    result = list(set(result))
    # Clean up temporary files
    KhmerSegment.remove_file(input_file)
    KhmerSegment.remove_file(output_file_w)
    KhmerSegment.remove_file(output_file_c)

    return result



import ntpath
import numpy as np

from khmerml.algorithms.decisiontree.tree import Tree
from khmerml.utils.file_util import FileUtil
from khmerml.algorithms.base import Base


class DecisionTree(Base, Tree):
  """
    DecisionTree class
  """

  def __init__(self, **kwargs):
    self.kwargs = kwargs
    self.train_model = {}
    self.predictions = []
    self.criterion = kwargs['criterion']
    self.prune = kwargs['prune']
    self.min_criterion = kwargs['min_criterion']
    Base.__init__(self, **kwargs)
    Tree.__init__(self, **kwargs)

  def load_model(self):
    """
      Load train model from file
    """
    try:
      head, tail = ntpath.split(self.kwargs['train_model'])
      model = FileUtil.load_model(head+ '/decision_tree_' +tail)
      self.__dict__.update(model)
    except IOError as error:
      raise Exception(error)
    return self

  def save_model(self, model):
    """
      Load train model from file
    """
    try:
      head, tail = ntpath.split(self.kwargs['train_model'])
      FileUtil.save_model(head+ '/decision_tree_' +tail, model.__dict__)
    except IOError as error:
      print(error)

  def train(self, dataset):
    """
      Train model
    """
    y_train = dataset[:, -1]
    X_train = np.delete(dataset, -1, 1)
    # Start constructing tree
    # self.build(X_train, y_train, self.criterion)
    self.build(X_train, y_train, self.criterion)
    self.save_model(self)
    # self.show_tree(10,'he')
    return self

  def predict(self, model, test_dataset):
    """
      Make prediction
    """
    # Get train and test label
    # Delete last column (label) from array
    X_test = np.delete(test_dataset, -1, 1)
    return np.array([self._predict(feature) for feature in X_test])



"""
  Machine learning abstract class
"""
from abc import ABC, abstractmethod
class Base(ABC):
  """
    Machine learning abstract class
  """

  def __init__(self, **kwargs):
    self.kwargs = kwargs

  @abstractmethod
  def load_model(self):
    """
      load dataset extraction
    """
    pass

  @abstractmethod
  def save_model(self, model):
    """
      load dataset extraction
    """
    pass

  @abstractmethod
  def train(self, dataset):
    """
      load dataset extraction
    """
    pass

  @abstractmethod
  def predict(self, model, test_dataset):
    """
      load dataset extraction
    """
    pass




import numpy as np
from numpy import exp

class NeuralNetworkBase(object):
  """
    NeuralNetwork Base class is use for calculation in
    Naive Bayes Algorithm
  """

  def __init__(self, **kwargs):
    """
      makes the random numbers predictable
      With the seed reset (every time), the same set of numbers will appear every time.
      If the random seed is not reset, different numbers appear with every invocation:
      random.seed(random_state)
      Set float precision
    """
    np.set_printoptions(precision=8)
    self.kwargs = kwargs


  def label_to_matrix(self, label_vector):
    """
      Transform labels to row-based vector
    """
    length = len(label_vector)
    n_label = len(set(label_vector))
    unique_labels = np.unique(label_vector)
    # Create matrix zeros m row x n columns
    label_matrix = np.zeros((length, n_label))
    # Replace value by 1 at position (i, label_vector(j))
    for i, value in enumerate(label_vector):
      #value = int(round(value))
      index = np.where(unique_labels == value)[0]
      label_matrix[i][index] = np.float(1.0)
    return label_matrix

  def make_weight_matrix(self, n_features, n_labels, hidden_layer_sizes):
    """
      Make random weight in form of matrix for each layer
    """
    layers = []
    change_output_layouts = []
    n_inputs = 0

    for _, n_output_nodes in enumerate(hidden_layer_sizes):
      #if index  is not layers:
      if n_inputs == 0:
          n_inputs = n_features

      # Create matrix of
      # n_inputs_neuron as row and n_neurons as column
      # Random value in interal [0, 1)

      # create randomized weights
      # use scheme from Efficient Backprop by LeCun 1998 to initialize weights for hidden layer
      input_range = 1.0 / n_inputs ** (1/2)
      synaptic_weight = np.random.normal(loc=0, scale=input_range, size=(n_inputs, n_output_nodes))

      # create arrays of 0 for changes
      # this is essentially an array of temporary values that gets updated at each iteration
      # based on how much the weights need to change in the following iteration
      change_output = np.zeros((n_inputs, n_output_nodes))
      change_output_layouts.append(change_output)

      # n_output_nodes is input for next layer
      n_inputs = n_output_nodes

      # Add layer to list
      synaptic_weight = np.array(synaptic_weight)
      layers.append(synaptic_weight)

    # Add layer to list
    # Output layer
    synaptic_weight = np.random.uniform(size=(n_inputs, n_labels)) / np.sqrt(n_inputs)
    layers.append(synaptic_weight)

    # create arrays of 0 for changes
    # this is essentially an array of temporary values that gets updated at each iteration
    # based on how much the weights need to change in the following iteration
    change_output = np.zeros((n_inputs, n_labels))
    change_output_layouts.append(change_output)
    return layers, change_output_layouts


  def activation_relu(self, X):
    """
      Compute the rectified linear unit function inplace.
      Parameters
      ----------
      X : {array-like, sparse matrix}, shape (n_samples, n_features)
          The input data.
      Returns
      -------
      X_new : {array-like, sparse matrix}, shape (n_samples, n_features)
      The transformed data.
    """
    np.clip(X, 0, np.finfo(X.dtype).max, out=X)
    return X


  def derivative_relu(self, X, delta):
    """
      Apply the derivative of the relu function.
      It exploits the fact that the derivative is a simple function of the output
      value from rectified linear units activation function.
      Parameters
      ----------
      Z : {array-like, sparse matrix}, shape (n_samples, n_features)
          The data which was output from the rectified linear units activation
          function during the forward pass.
      delta : {array-like}, shape (n_samples, n_features)
          The backpropagated error signal to be modified inplace.
    """
    delta[X == 0] = 0
    return delta

  def activation_tanh(self, X):
    """
      tanh is a little nicer than the standard 1/(1+e^-x)
    """
    return np.tanh(X, out=X)

  def derivative_tanh(self, X, delta):
    """
      Apply the derivative of the hyperbolic tanh function.
      It exploits the fact that the derivative is a simple function of the output
      value from hyperbolic tangent.
      Parameters
      ----------
      Z : {array-like, sparse matrix}, shape (n_samples, n_features)
          The data which was output from the hyperbolic tangent activation
          function during the forward pass.
      delta : {array-like}, shape (n_samples, n_features)
          The backpropagated error signal to be modified inplace.
    """
    #delta *= (1 - X ** 2)
    #delta *= 1.1493*(1 - ((2/3)*X) ** 2)
    delta *= (1 - X ** 2)
    return delta


  def activation_sigmoid_exp(self, X):
    """
      The Sigmoid function, which describes an S shaped curve.
      We pass the weighted sum of the inputs through this function to
      normalise them between 0 and 1.
    """
    return 1 / (1 + exp(-X))

  def derivative_sigmoid(self, X, delta):
    """
      Apply the derivative of the logistic sigmoid function.
      It exploits the fact that the derivative is a simple function of the output
      value from logistic function.
      Parameters
      ----------
      Z : {array-like, sparse matrix}, shape (n_samples, n_features)
          The data which was output from the logistic activation function during
          the forward pass.
      delta : {array-like}, shape (n_samples, n_features)
          The backpropagated error signal to be modified inplace.
    """
    delta *= X
    delta *= (1 - X)
    return delta



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: "KhmerML" <khmerml@aicambodia.com>
#
from __future__ import division, print_function
from setuptools import setup, find_packages
DOCLINES = (__doc__ or '').split("\n")
import os
import sys
import subprocess

if sys.version_info[:2] < (2, 7) or (3, 0) <= sys.version_info[:2] < (3, 4):
    raise RuntimeError("Python version 2.7 or >= 3.4 required.")

if sys.version_info[0] >= 3:
    import builtins
else:
    import __builtin__ as builtins


CLASSIFIERS = """\
Development Status :: 5 - Production/Stable
Intended Audience :: Science/Research
Intended Audience :: Developers
License :: OSI Approved :: GNU General Public License v3 (GPLv3)
Programming Language :: Python
Programming Language :: Python :: 2
Programming Language :: Python :: 2.7
Programming Language :: Python :: 3
Programming Language :: Python :: 3.4
Programming Language :: Python :: 3.5
Programming Language :: Python :: 3.6
Programming Language :: Python :: 3.7
Topic :: Software Development
Topic :: Scientific/Engineering
Operating System :: Microsoft :: Windows
Operating System :: POSIX
Operating System :: Unix
Operating System :: MacOS
"""

DESCRIPTION = """khmerML: opensource module for machine learning"""

LONG_DESCRIPTION = """khmerML is an opensource Python module for machine learning that consists of machine learning algorithms building from scratch focus on solving complex problems in Cambodia society. It has developed by slash research team and it will be able to contribute by anybody that willing to share their research.

khmerML also is a way to encourage Cambodian tech engineers to start learning Machine Learning."""

MAJOR               = 0
MINOR               = 0
MICRO               = 12
ISRELEASED          = True
VERSION             = '%d.%d.%d' % (MAJOR, MINOR, MICRO)


# Return the git revision as a string
def git_version():
    def _minimal_ext_cmd(cmd):
        # construct minimal environment
        env = {}
        for k in ['SYSTEMROOT', 'PATH', 'HOME']:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        out = subprocess.Popen(cmd, stdout=subprocess.PIPE, env=env).communicate()[0]
        return out

    try:
        out = _minimal_ext_cmd(['git', 'rev-parse', 'HEAD'])
        GIT_REVISION = out.strip().decode('ascii')
    except OSError:
        GIT_REVISION = "Unknown"

    return GIT_REVISION

# BEFORE importing setuptools, remove MANIFEST. Otherwise it may not be
# properly updated when the contents of directories change (true for distutils,
# not sure about setuptools).
if os.path.exists('MANIFEST'):
    os.remove('MANIFEST')

# This is a bit hackish: we are setting a global variable so that the main
# khmerml __init__ can detect if it is being loaded by the setup routine, to
# avoid attempting to load components that aren't built yet.  While ugly, it's
# a lot more robust than what was previously being used.
builtins.__KHMERML_SETUP__ = True


def get_version_info():
    # Adding the git rev number needs to be done inside write_version_py(),
    # otherwise the import of khmerml.version messes up the build under Python 3.
    FULLVERSION = VERSION
    if os.path.exists('.git'):
        GIT_REVISION = git_version()
    elif os.path.exists('khmerml/version.py'):
        # must be a source distribution, use existing version file
        try:
            from khmerml.version import git_revision as GIT_REVISION
        except ImportError:
            raise ImportError("Unable to import git_revision. Try removing " \
                              "khmerml/version.py and the build directory " \
                              "before building.")
    else:
        GIT_REVISION = "Unknown"

    if not ISRELEASED:
        FULLVERSION += '.dev0+' + GIT_REVISION[:7]

    return FULLVERSION, GIT_REVISION


def write_version_py(filename='khmerml/version.py'):
    cnt = """
# THIS FILE IS GENERATED FROM khmerml SETUP.PY
#
# To compare versions robustly, use `khmerml.lib.khmermlVersion`
short_version = '%(version)s'
version = '%(version)s'
full_version = '%(full_version)s'
git_revision = '%(git_revision)s'
release = %(isrelease)s
if not release:
    version = full_version
"""
    FULLVERSION, GIT_REVISION = get_version_info()

    a = open(filename, 'w')
    try:
        a.write(cnt % {'version': VERSION,
                       'full_version': FULLVERSION,
                       'git_revision': GIT_REVISION,
                       'isrelease': str(ISRELEASED)})
    finally:
        a.close()



default_setup = dict(
    name='khmerml',
    version= get_version_info()[0],
    description = DESCRIPTION,
    long_description = LONG_DESCRIPTION,
    provides=['khmerml'],
    install_requires=['numpy==1.14.5', 'nltk>=3.2.4'],
    requires=['numpy', 'nltk'],
    packages=find_packages(),
    classifiers=[_f for _f in CLASSIFIERS.split('\n') if _f],
    platforms = ["Windows", "Linux", "Solaris", "Mac OS-X", "Unix"],
    author='KhmerML',
    author_email='kevin@slash.co',
)

setup(**default_setup)

# Install nltk data such as corpora, toy grammars, trained models, etc.
# import subprocess
# subprocess.Popen(['sh','nltk_data.sh'])




import logging

class Log(object):

  def __init__(self, *args, **kwargs):
    self.kwargs = kwargs
    self.filename = args[0]

  def log(self,data):
    logging.basicConfig(filename= self.filename,level=logging.DEBUG)
    logging.debug(data)





import numpy as np
import math
from decimal import Decimal
from functools import reduce

class BayesBase(object):
  """
    BayesBase class is use for calculation in
    Naive Bayes Algorithm
  """

  def __init__(self, **kwargs):
    self.kwargs = kwargs
    self._train_model = {}
    self.predictions = []


  def count_classes_occurrence(self, X_train):
    """
      Count class occurences from list data set by class
    """

    _, classes_occurrence = np.unique(X_train, return_counts=True)
    return classes_occurrence


  def calculate_priori(self, X_train):
    """
      Calculate Priori Probability
    """

    # Count class occurences from X_train
    classes, classes_occurrence = np.unique(X_train[:, -1], return_counts=True)
    total_occurences = np.sum(classes_occurrence)

    # Calculate probability per class
    prioris = {}
    for index, label in enumerate(classes):
      prioris[label] = classes_occurrence[index] / total_occurences

    return prioris


  def calculate_likelihood(self, X_train):
    """
      Calculate likelihoods
    """

    # Unique class
    unique_class = np.unique(X_train[:,-1])

    likelihoods = {}
    for _, label in enumerate(unique_class):

      # Find all rows where its label equal to search label
      X_subset = X_train[X_train[:, -1] == label]
      # Delete lable from dataset
      X_subset = np.delete(X_subset, -1, 1)
      # Count total frequency of all features
      total_frequency = np.sum(X_subset)
      # Count number of features
      total_features = X_subset.shape[1]

      # Count number of frequency of each and every feature
      feature_fequency = [np.sum(feature) for feature in X_subset.T]
      # Calculate likelihood of each feature
      probabilities = [(1+ f_count)/(total_features + total_frequency) \
                  for f_count in feature_fequency]

      # Push likelihood to list
      likelihoods[label] = probabilities

    return likelihoods

  def calculate_posteriori(self, model, test_vector):
    """
      Calculate the porbability of all classes
      one class at a time.
    """
    best_posteriori, best_label = -1, None

    for label, priori_likelihood in model.items():

      # Get priori (index=0)
      priori = Decimal(priori_likelihood[0])
      # Get likelihood (index=1)
      likelihood_feature = priori_likelihood[1]

      _likelihoods = list(map(lambda x, y: math.pow(x, y), \
                      likelihood_feature, test_vector))

      likelihood = reduce(lambda x, y: Decimal(x) * Decimal(y), _likelihoods)

      # Calculate posteriori
      posteriori = priori * Decimal(likelihood)

      # Check for the best posteriori
      if best_label is None or posteriori > best_posteriori:
        best_posteriori = posteriori
        best_label = label

    return best_posteriori, best_label



"""
"""
import csv
import os
import pickle
import zipfile
import shutil

from numpy import genfromtxt

class FileUtil(object):
  """
    File Utilities for read and write file
  """

  @staticmethod
  def read_file(filename, file_extension):
    """
      Read file
      return: dataset as list
    """
    _data_matrix = []
    if file_extension == 'CSV':
      _data_matrix = FileUtil.load_csv(filename)

    return _data_matrix

  @staticmethod
  def load_csv(filename, use_numpy=True):
    """
      Read data from csv file using python or numpy lib
    """
    if not use_numpy:
      return FileUtil.load_csv_py(filename)
    else:
      return FileUtil.load_csv_np(filename)

  @staticmethod
  def load_csv_py(filename):
    """
      Read data from csv file.
    """
    file_obj = open(filename, "rU")
    file_csv = csv.reader(file_obj)
    _dataset = list(file_csv)
    for index, subset in enumerate(_dataset):
      _dataset[index] = [float(x) for x in subset]
    file_obj.close()
    return _dataset

  @staticmethod
  def load_csv_np(filename):
    """
      Load data from a text file
      When spaces are used as delimiters, or when no delimiter has been given as input\
      , there should not be any missing data between two fields.
      When the variables are named (either by a flexible dtype or with names\
      , there must not be any header in the file (else a ValueError exception is raised).
      Individual values are not stripped of spaces by default. When using a custom converter
      , make sure the function does remove spaces
    """
    # Data read from the text file. If usemask is True, this is a masked array.
    _dataset = genfromtxt(filename, delimiter=',')

    return _dataset

  @staticmethod
  def dataset_path(config, filename):
    """
      path to file
    """
    path = os.path.join(os.getcwd(), config['dataset'], filename)
    return path

  @staticmethod
  def path_to_file(dirname, filename):
    """
      path to file
    """

    path = os.path.join(os.getcwd(), dirname, filename)
    return path

  @staticmethod
  def join_path(dirname):
    """
      path to file
    """
    path = os.path.join(os.getcwd(), dirname)
    return path

  @staticmethod
  def save_pickle_dataset(pickle_filename, dataset):
    """
      The pickle module implements binary protocols
      for serializing and de-serializing a Python object structure.
    """
    path_to_pickle = os.path.join(os.getcwd(), pickle_filename)
    with open(path_to_pickle, 'wb') as handle:
      pickle.dump(dataset, handle, protocol=pickle.HIGHEST_PROTOCOL)

    return True

  @staticmethod
  def save_model(file, model):
    """
      The pickle module implements binary protocols
      for serializing and de-serializing a Python object structure.
    """
    path_to_pickle = os.path.join(os.getcwd(), file)
    try:
      FileUtil.create_folder(FileUtil.get_folder_path(path_to_pickle))
      with open(path_to_pickle, 'wb') as handle:
        pickle.dump(model, handle, protocol=pickle.HIGHEST_PROTOCOL)
    except pickle.PickleError as error:
      raise Exception(error)
    else:
        return True

  @staticmethod
  def load_model(file):
    """
      Read .pickle file
    """
    path_to_pickle = os.path.join(os.getcwd(), file)
    try:
      with open(path_to_pickle, 'rb') as handle:
        model = pickle.load(handle)
    except pickle.UnpicklingError as error:
      raise Exception(error)
    else:
      return model

  @staticmethod
  def load_pickle(pickle_filename):
    """
      Read .pickle file
    """
    path_to_pickle = os.path.join(os.getcwd(), pickle_filename)
    try:
      with open(path_to_pickle, 'rb') as handle:
        model = pickle.load(handle)
    except pickle.UnpicklingError as error:
      raise Exception(error)
    else:
      return model

  @staticmethod
  def extract_zipfile(path_to_zipefile, dest_path):
    """
      Extract zipfile sent from client
      Store temporarily in server
    """
    try:
      with zipfile.ZipFile(path_to_zipefile) as opened_rar:
        opened_rar.extractall(dest_path)
    except OSError as error:
        raise Exception(error)
    else:
        return True

  @staticmethod
  def move_file(source, destination):
    """
      This function is used to move file from source to destination
    """
    try:
      shutil.move(source, destination)
    except OSError as error:
      print('error %s' % error)
      return False
    return True

  @staticmethod
  def remove_file(path, ignore_errors=False, onerror=None):
    """
      Delete an entire directory tree; path must point to a directory \
      (but not a symbolic link to a directory). If ignore_errors is true, errors \
      resulting from failed removals will be ignored; if false or omitted, \
      such errors are handled by calling a handler specified by onerror or, \
      if that is omitted, they raise an exception.
    """

    ## check if a file exists on disk ##
    ## if exists, delete it else show message on screen ##
    if os.path.exists(path):
      if os.path.isfile(path):
        try:
          os.remove(path)
        except OSError as error:
          print("Error: %s - %s." % (error.filename, error.strerror))
      else:
        try:
          shutil.rmtree(path, ignore_errors, onerror)
        except OSError as error:
          print('error %s' % error)
          return False
    else:
      print("Sorry, I can not find %s file." % path)
      return False
    return True

  @staticmethod
  def _print(*args):
    try:
      # print(args)
      pass
    except UnicodeEncodeError:
      print('cannot print')

  @staticmethod
  def create_folder(folder):
    if not os.path.exists(folder):
      os.makedirs(folder)

  @staticmethod
  def get_folder_path(file):
    return os.path.dirname(file)
